@@ -1,3 +1,5 @@
+# -*- coding: utf-8 -*-
+
 ## Create and compare lists of files/objects
 ## Author: Michal Ludvig <michal@logix.cz>
 ##         http://www.logix.cz/michal
@@ -12,17 +14,76 @@ from Utils import *
 from Exceptions import ParameterError
 from HashCache import HashCache
 
-from logging import debug, info, warning, error
+from logging import debug, info, warning
 
 import os
 import sys
 import glob
-import copy
 import re
 import errno
 
 __all__ = ["fetch_local_list", "fetch_remote_list", "compare_filelists"]
 
+def _os_walk_unicode(top):
+    '''
+    Reimplementation of python's os.walk to nicely support unicode in input as in output.
+    '''
+    try:
+        names = os.listdir(deunicodise(top))
+    except:
+        return
+
+    dirs, nondirs = [], []
+    for name in names:
+        name = unicodise(name)
+        if os.path.isdir(deunicodise(os.path.join(top, name))):
+            if not handle_exclude_include_walk_dir(top, name):
+                dirs.append(name)
+        else:
+            nondirs.append(name)
+
+    yield top, dirs, nondirs
+    for name in dirs:
+        new_path = os.path.join(top, name)
+        if not os.path.islink(deunicodise(new_path)):
+            for x in _os_walk_unicode(new_path):
+                yield x
+
+def handle_exclude_include_walk_dir(root, dirname):
+    '''
+    Should this root/dirname directory be excluded? (otherwise included by default)
+    Exclude dir matches in the current directory
+    This prevents us from recursing down trees we know we want to ignore
+    return True for including, and False for excluding
+    '''
+    cfg = Config()
+    d = os.path.join(root, dirname, '')
+    debug(u"CHECK: %r" % d)
+    excluded = False
+    for r in cfg.exclude:
+        # python versions end their patterns (from globs) differently, test for both styles.
+        if not (r.pattern.endswith(u'\\/$') or r.pattern.endswith(u'\\/\\Z(?ms)')): continue # we only check for directory patterns here
+        if r.search(d):
+            excluded = True
+            debug(u"EXCL-MATCH: '%s'" % (cfg.debug_exclude[r]))
+            break
+    if excluded:
+        ## No need to check for --include if not excluded
+        for r in cfg.include:
+            # python versions end their patterns (from globs) differently, test for both styles.
+            if not (r.pattern.endswith(u'\\/$') or r.pattern.endswith(u'\\/\\Z(?ms)')): continue # we only check for directory patterns here
+            debug(u"INCL-TEST: %s ~ %s" % (d, r.pattern))
+            if r.search(d):
+                excluded = False
+                debug(u"INCL-MATCH: '%s'" % (cfg.debug_include[r]))
+                break
+    if excluded:
+        ## Still excluded - ok, action it
+        debug(u"EXCLUDE: %r" % d)
+    else:
+        debug(u"PASS: %r" % d)
+    return excluded
+
 def _fswalk_follow_symlinks(path):
     '''
     Walk filesystem, following symbolic links (but without recursion), on python2.4 and later
@@ -30,23 +91,21 @@ def _fswalk_follow_symlinks(path):
     If a symlink directory loop is detected, emit a warning and skip.
     E.g.: dir1/dir2/sym-dir -> ../dir2
     '''
-    assert os.path.isdir(path) # only designed for directory argument
+    assert os.path.isdir(deunicodise(path)) # only designed for directory argument
     walkdirs = set([path])
-    for dirpath, dirnames, filenames in os.walk(path):
-        handle_exclude_include_walk(dirpath, dirnames, [])
-        real_dirpath = os.path.realpath(dirpath)
+    for dirpath, dirnames, filenames in _os_walk_unicode(path):
+        real_dirpath = unicodise(os.path.realpath(deunicodise(dirpath)))
         for dirname in dirnames:
             current = os.path.join(dirpath, dirname)
-            real_current = os.path.realpath(current)
-            if os.path.islink(current):
+            real_current = unicodise(os.path.realpath(deunicodise(current)))
+            if os.path.islink(deunicodise(current)):
                 if (real_dirpath == real_current or
                     real_dirpath.startswith(real_current + os.path.sep)):
                     warning("Skipping recursively symlinked directory %s" % dirname)
                 else:
                     walkdirs.add(current)
     for walkdir in walkdirs:
-        for dirpath, dirnames, filenames in os.walk(walkdir):
-            handle_exclude_include_walk(dirpath, dirnames, [])
+        for dirpath, dirnames, filenames in _os_walk_unicode(walkdir):
             yield (dirpath, dirnames, filenames)
 
 def _fswalk_no_symlinks(path):
@@ -55,8 +114,7 @@ def _fswalk_no_symlinks(path):
 
     path (str) is the root of the directory tree to walk
     '''
-    for dirpath, dirnames, filenames in os.walk(path):
-        handle_exclude_include_walk(dirpath, dirnames, filenames)
+    for dirpath, dirnames, filenames in _os_walk_unicode(path):
         yield (dirpath, dirnames, filenames)
 
 def filter_exclude_include(src_list):
@@ -88,40 +146,6 @@ def filter_exclude_include(src_list):
             debug(u"PASS: %r" % (file))
     return src_list, exclude_list
 
-def handle_exclude_include_walk(root, dirs, files):
-    cfg = Config()
-    copydirs = copy.copy(dirs)
-    # exclude dir matches in the current directory
-    # this prevents us from recursing down trees we know we want to ignore
-    for x in copydirs:
-        d = os.path.join(root, x, '')
-        debug(u"CHECK: %r" % d)
-        excluded = False
-        for r in cfg.exclude:
-            # python versions end their patterns (from globs) differently, test for both styles.
-            if not (r.pattern.endswith(u'\\/$') or r.pattern.endswith(u'\\/\\Z(?ms)')): continue # we only check for directory patterns here
-            if r.search(d):
-                excluded = True
-                debug(u"EXCL-MATCH: '%s'" % (cfg.debug_exclude[r]))
-                break
-        if excluded:
-            ## No need to check for --include if not excluded
-            for r in cfg.include:
-                # python versions end their patterns (from globs) differently, test for both styles.
-                if not (r.pattern.endswith(u'\\/$') or r.pattern.endswith(u'\\/\\Z(?ms)')): continue # we only check for directory patterns here
-                debug(u"INCL-TEST: %s ~ %s" % (d, r.pattern))
-                if r.search(d):
-                    excluded = False
-                    debug(u"INCL-MATCH: '%s'" % (cfg.debug_include[r]))
-                    break
-        if excluded:
-            ## Still excluded - ok, action it
-            debug(u"EXCLUDE: %r" % d)
-            dirs.remove(x)
-            continue
-        else:
-            debug(u"PASS: %r" % (d))
-
 
 def _get_filelist_from_file(cfg, local_path):
     def _append(d, key, value):
@@ -136,16 +160,16 @@ def _get_filelist_from_file(cfg, local_path):
             f = sys.stdin
         else:
             try:
-                f = open(fname, 'r')
+                f = open(deunicodise(fname), 'r')
             except IOError, e:
                 warning(u"--files-from input file %s could not be opened for reading (%s), skipping." % (fname, e.strerror))
                 continue
 
         for line in f:
-            line = line.strip()
+            line = unicodise(line).strip()
             line = os.path.normpath(os.path.join(local_path, line))
-            dirname = os.path.dirname(line)
-            basename = os.path.basename(line)
+            dirname = unicodise(os.path.dirname(deunicodise(line)))
+            basename = unicodise(os.path.basename(deunicodise(line)))
             _append(filelist, dirname, basename)
         if f != sys.stdin:
             f.close()
@@ -164,6 +188,7 @@ def fetch_local_list(args, is_src = False, recursive = None):
 
     def _fetch_local_list_info(loc_list):
         len_loc_list = len(loc_list)
+        total_size = 0
         info(u"Running stat() and reading/calculating MD5 values on %d files, this may take some time..." % len_loc_list)
         counter = 0
         for relative_file in loc_list:
@@ -175,7 +200,7 @@ def fetch_local_list(args, is_src = False, recursive = None):
 
             full_name = loc_list[relative_file]['full_name']
             try:
-                sr = os.stat_result(os.stat(full_name))
+                sr = os.stat_result(os.stat(deunicodise(full_name)))
             except OSError, e:
                 if e.errno == errno.ENOENT:
                     # file was removed async to us getting the list
@@ -192,6 +217,7 @@ def fetch_local_list(args, is_src = False, recursive = None):
                 'sr': sr # save it all, may need it in preserve_attrs_list
                 ## TODO: Possibly more to save here...
             })
+            total_size += sr.st_size
             if 'md5' in cfg.sync_checks:
                 md5 = cache.md5(sr.st_dev, sr.st_ino, sr.st_mtime, sr.st_size)
                 if md5 is None:
@@ -201,12 +227,13 @@ def fetch_local_list(args, is_src = False, recursive = None):
                             continue
                         cache.add(sr.st_dev, sr.st_ino, sr.st_mtime, sr.st_size, md5)
                 loc_list.record_hardlink(relative_file, sr.st_dev, sr.st_ino, md5, sr.st_size)
+        return total_size
 
 
     def _get_filelist_local(loc_list, local_uri, cache):
         info(u"Compiling list of local files...")
 
-        if deunicodise(local_uri.basename()) == "-":
+        if local_uri.basename() == "-":
             try:
                 uid = os.geteuid()
                 gid = os.getegid()
@@ -214,7 +241,6 @@ def fetch_local_list(args, is_src = False, recursive = None):
                 uid = 0
                 gid = 0
             loc_list["-"] = {
-                'full_name_unicode' : '-',
                 'full_name' : '-',
                 'size' : -1,
                 'mtime' : -1,
@@ -225,8 +251,8 @@ def fetch_local_list(args, is_src = False, recursive = None):
             }
             return loc_list, True
         if local_uri.isdir():
-            local_base = deunicodise(local_uri.basename())
-            local_path = deunicodise(local_uri.path())
+            local_base = local_uri.basename()
+            local_path = local_uri.path()
             if is_src and len(cfg.files_from):
                 filelist = _get_filelist_from_file(cfg, local_path)
                 single_file = False
@@ -238,19 +264,22 @@ def fetch_local_list(args, is_src = False, recursive = None):
                 single_file = False
         else:
             local_base = ""
-            local_path = deunicodise(local_uri.dirname())
-            filelist = [( local_path, [], [deunicodise(local_uri.basename())] )]
+            local_path = local_uri.dirname()
+            filelist = [( local_path, [], [local_uri.basename()] )]
             single_file = True
         for root, dirs, files in filelist:
             rel_root = root.replace(local_path, local_base, 1)
             for f in files:
                 full_name = os.path.join(root, f)
-                if not os.path.isfile(full_name):
+                if not os.path.isfile(deunicodise(full_name)):
+                    if os.path.exists(deunicodise(full_name)):
+                        warning(u"Skipping over non regular file: %s" % full_name)
                     continue
-                if os.path.islink(full_name):
-                                    if not cfg.follow_symlinks:
-                                            continue
-                relative_file = unicodise(os.path.join(rel_root, f))
+                if os.path.islink(deunicodise(full_name)):
+                    if not cfg.follow_symlinks:
+                        warning(u"Skipping over symbolic link: %s" % full_name)
+                        continue
+                relative_file = os.path.join(rel_root, f)
                 if os.path.sep != "/":
                     # Convert non-unix dir separators to '/'
                     relative_file = "/".join(relative_file.split(os.path.sep))
@@ -259,7 +288,6 @@ def fetch_local_list(args, is_src = False, recursive = None):
                 if relative_file.startswith('./'):
                     relative_file = relative_file[2:]
                 loc_list[relative_file] = {
-                    'full_name_unicode' : unicodise(full_name),
                     'full_name' : full_name,
                 }
 
@@ -290,7 +318,7 @@ def fetch_local_list(args, is_src = False, recursive = None):
     local_list = FileDict(ignore_case = False)
     single_file = False
 
-    if type(args) not in (list, tuple):
+    if type(args) not in (list, tuple, set):
         args = [args]
 
     if recursive == None:
@@ -316,17 +344,20 @@ def fetch_local_list(args, is_src = False, recursive = None):
         single_file = False
 
     local_list, exclude_list = filter_exclude_include(local_list)
-    _fetch_local_list_info(local_list)
+    total_size = _fetch_local_list_info(local_list)
     _maintain_cache(cache, local_list)
-    return local_list, single_file, exclude_list
+    return local_list, single_file, exclude_list, total_size
 
 def fetch_remote_list(args, require_attribs = False, recursive = None, uri_params = {}):
     def _get_remote_attribs(uri, remote_item):
         response = S3(cfg).object_info(uri)
+        if not response.get('headers'):
+            return
+
         remote_item.update({
         'size': int(response['headers']['content-length']),
         'md5': response['headers']['etag'].strip('"\''),
-        'timestamp' : dateRFC822toUnix(response['headers']['date'])
+        'timestamp' : dateRFC822toUnix(response['headers']['last-modified'])
         })
         try:
             md5 = response['s3cmd-attrs']['md5']
@@ -354,6 +385,8 @@ def fetch_remote_list(args, require_attribs = False, recursive = None, uri_param
         info(u"Retrieving list of remote files for %s ..." % remote_uri)
         empty_fname_re = re.compile(r'\A\s*\Z')
 
+        total_size = 0
+
         s3 = S3(Config())
         response = s3.bucket_list(remote_uri.bucket(), prefix = remote_uri.object(),
                                   recursive = recursive, uri_params = uri_params)
@@ -362,14 +395,14 @@ def fetch_remote_list(args, require_attribs = False, recursive = None, uri_param
         remote_uri_original = remote_uri
         if rem_base != '' and rem_base[-1] != '/':
             rem_base = rem_base[:rem_base.rfind('/')+1]
-            remote_uri = S3Uri("s3://%s/%s" % (remote_uri.bucket(), rem_base))
+            remote_uri = S3Uri(u"s3://%s/%s" % (remote_uri.bucket(), rem_base))
         rem_base_len = len(rem_base)
         rem_list = FileDict(ignore_case = False)
         break_now = False
         for object in response['list']:
             if object['Key'] == rem_base_original and object['Key'][-1] != "/":
                 ## We asked for one file and we got that file :-)
-                key = os.path.basename(object['Key'])
+                key = unicodise(os.path.basename(deunicodise(object['Key'])))
                 object_uri_str = remote_uri_original.uri()
                 break_now = True
                 rem_list = FileDict(ignore_case = False)   ## Remove whatever has already been put to rem_list
@@ -394,15 +427,16 @@ def fetch_remote_list(args, require_attribs = False, recursive = None, uri_param
                 _get_remote_attribs(S3Uri(object_uri_str), rem_list[key])
             md5 = rem_list[key]['md5']
             rem_list.record_md5(key, md5)
+            total_size += int(object['Size'])
             if break_now:
                 break
-        return rem_list
+        return rem_list, total_size
 
     cfg = Config()
     remote_uris = []
     remote_list = FileDict(ignore_case = False)
 
-    if type(args) not in (list, tuple):
+    if type(args) not in (list, tuple, set):
         args = [args]
 
     if recursive == None:
@@ -414,15 +448,18 @@ def fetch_remote_list(args, require_attribs = False, recursive = None, uri_param
             raise ParameterError("Expecting S3 URI instead of '%s'" % arg)
         remote_uris.append(uri)
 
+    total_size = 0
+
     if recursive:
         for uri in remote_uris:
-            objectlist = _get_filelist_remote(uri, recursive = True)
+            objectlist, tmp_total_size = _get_filelist_remote(uri, recursive = True)
+            total_size += tmp_total_size
             for key in objectlist:
                 remote_list[key] = objectlist[key]
                 remote_list.record_md5(key, objectlist.get_md5(key))
     else:
         for uri in remote_uris:
-            uri_str = unicode(uri)
+            uri_str = uri.uri()
             ## Wildcards used in remote URI?
             ## If yes we'll need a bucket listing...
             wildcard_split_result = re.split("\*|\?", uri_str, maxsplit=1)
@@ -431,19 +468,20 @@ def fetch_remote_list(args, require_attribs = False, recursive = None, uri_param
                 ## Only request recursive listing if the 'rest' of the URI,
                 ## i.e. the part after first wildcard, contains '/'
                 need_recursion = '/' in rest
-                objectlist = _get_filelist_remote(S3Uri(prefix), recursive = need_recursion)
+                objectlist, tmp_total_size = _get_filelist_remote(S3Uri(prefix), recursive = need_recursion)
+                total_size += tmp_total_size
                 for key in objectlist:
                     ## Check whether the 'key' matches the requested wildcards
                     if glob.fnmatch.fnmatch(objectlist[key]['object_uri_str'], uri_str):
                         remote_list[key] = objectlist[key]
             else:
                 ## No wildcards - simply append the given URI to the list
-                key = os.path.basename(uri.object())
+                key = unicodise(os.path.basename(deunicodise(uri.object())))
                 if not key:
                     raise ParameterError(u"Expecting S3 URI with a filename or --recursive: %s" % uri.uri())
                 remote_item = {
                     'base_uri': uri,
-                    'object_uri_str': unicode(uri),
+                    'object_uri_str': uri.uri(),
                     'object_key': uri.object()
                 }
                 if require_attribs:
@@ -453,12 +491,13 @@ def fetch_remote_list(args, require_attribs = False, recursive = None, uri_param
                 md5 = remote_item.get('md5')
                 if md5:
                     remote_list.record_md5(key, md5)
+                total_size += remote_item.get('size', 0)
 
     remote_list, exclude_list = filter_exclude_include(remote_list)
-    return remote_list, exclude_list
+    return remote_list, exclude_list, total_size
 
 
-def compare_filelists(src_list, dst_list, src_remote, dst_remote, delay_updates = False):
+def compare_filelists(src_list, dst_list, src_remote, dst_remote):
     def __direction_str(is_remote):
         return is_remote and "remote" or "local"
 
@@ -487,7 +526,7 @@ def compare_filelists(src_list, dst_list, src_remote, dst_remote, delay_updates
             try:
                 src_md5 = src_list.get_md5(file)
                 dst_md5 = dst_list.get_md5(file)
-            except (IOError,OSError), e:
+            except (IOError,OSError):
                 # md5 sum verification failed - ignore that file altogether
                 debug(u"IGNR: %s (disappeared)" % (file))
                 warning(u"%s: file disappeared, ignoring." % (file))
@@ -527,7 +566,7 @@ def compare_filelists(src_list, dst_list, src_remote, dst_remote, delay_updates
 
             try:
                 same_file = _compare(src_list, dst_list, src_remote, dst_remote, relative_file)
-            except (IOError,OSError), e:
+            except (IOError,OSError):
                 debug(u"IGNR: %s (disappeared)" % (relative_file))
                 warning(u"%s: file disappeared, ignoring." % (relative_file))
                 del(src_list[relative_file])
